@@ -12,16 +12,19 @@ from fontTools.pens.recordingPen import RecordingPen
 from fontTools.pens.statisticsPen import StatisticsPen
 from fontTools.pens.momentsPen import OpenContourError
 from collections import OrderedDict
+import math
 import itertools
 import sys
 
+
 def _rot_list(l, k):
     """Rotate list by k items forward.  Ie. item at position 0 will be
     at position k in returned list.  Negative k is allowed."""
     n = len(l)
     k %= n
-    if not k: return l
-    return l[n-k:] + l[:n-k]
+    if not k:
+        return l
+    return l[n - k :] + l[: n - k]
 
 
 class PerContourPen(BasePen):
@@ -65,11 +68,10 @@ class PerContourOrComponentPen(PerContourPen):
 
 
 class RecordingPointPen(BasePen):
-
     def __init__(self):
         self.value = []
 
-    def beginPath(self, identifier = None, **kwargs):
+    def beginPath(self, identifier=None, **kwargs):
         pass
 
     def endPath(self) -> None:
@@ -89,6 +91,7 @@ def _vlen(vec):
         v += x * x
     return v
 
+
 def _complex_vlen(vec):
     v = 0
     for x in vec:
@@ -135,12 +138,13 @@ def min_cost_perfect_bipartite_matching(G):
     return best, best_cost
 
 
-def test(glyphsets, glyphs=None, names=None):
-
+def test(glyphsets, glyphs=None, names=None, ignore_missing=False):
     if names is None:
         names = glyphsets
     if glyphs is None:
-        glyphs = glyphsets[0].keys()
+        # `glyphs = glyphsets[0].keys()` is faster, certainly, but doesn't allow for sparse TTFs/OTFs given out of order
+        # ... risks the sparse master being the first one, and only processing a subset of the glyphs
+        glyphs = {g for glyphset in glyphsets for g in glyphset.keys()}
 
     hist = []
     problems = OrderedDict()
@@ -149,20 +153,22 @@ def test(glyphsets, glyphs=None, names=None):
         problems.setdefault(glyphname, []).append(problem)
 
     for glyph_name in glyphs:
-        # print()
-        # print(glyph_name)
-
         try:
+            m0idx = 0
             allVectors = []
             allNodeTypes = []
             allContourIsomorphisms = []
             for glyphset, name in zip(glyphsets, names):
-                # print('.', end='')
-                if glyph_name not in glyphset:
-                    add_problem(glyph_name, {"type": "missing", "master": name})
-                    continue
                 glyph = glyphset[glyph_name]
 
+                if glyph is None:
+                    if not ignore_missing:
+                        add_problem(glyph_name, {"type": "missing", "master": name})
+                    allNodeTypes.append(None)
+                    allVectors.append(None)
+                    allContourIsomorphisms.append(None)
+                    continue
+
                 perContourPen = PerContourOrComponentPen(
                     RecordingPen, glyphset=glyphset
                 )
@@ -180,7 +186,6 @@ def test(glyphsets, glyphs=None, names=None):
                 allVectors.append(contourVectors)
                 allContourIsomorphisms.append(contourIsomorphisms)
                 for ix, contour in enumerate(contourPens):
-
                     nodeVecs = tuple(instruction[0] for instruction in contour.value)
                     nodeTypes.append(nodeVecs)
 
@@ -193,7 +198,7 @@ def test(glyphsets, glyphs=None, names=None):
                             {"master": name, "contour": ix, "type": "open_path"},
                         )
                         continue
-                    size = abs(stats.area) ** 0.5 * 0.5
+                    size = math.sqrt(abs(stats.area)) * 0.5
                     vector = (
                         int(size),
                         int(stats.meanX),
@@ -206,10 +211,10 @@ def test(glyphsets, glyphs=None, names=None):
                     # print(vector)
 
                     # Check starting point
-                    if nodeVecs[0] == 'addComponent':
+                    if nodeVecs[0] == "addComponent":
                         continue
-                    assert nodeVecs[0] == 'moveTo'
-                    assert nodeVecs[-1] in ('closePath', 'endPath')
+                    assert nodeVecs[0] == "moveTo"
+                    assert nodeVecs[-1] in ("closePath", "endPath")
                     points = RecordingPointPen()
                     converter = SegmentToPointPen(points, False)
                     contour.replay(converter)
@@ -217,36 +222,47 @@ def test(glyphsets, glyphs=None, names=None):
                     # now check all rotations and mirror-rotations of the contour and build list of isomorphic
                     # possible starting points.
                     bits = 0
-                    for pt,b in points.value:
+                    for pt, b in points.value:
                         bits = (bits << 1) | b
                     n = len(points.value)
-                    mask = (1 << n ) - 1
+                    mask = (1 << n) - 1
                     isomorphisms = []
                     contourIsomorphisms.append(isomorphisms)
                     for i in range(n):
                         b = ((bits << i) & mask) | ((bits >> (n - i)))
                         if b == bits:
-                            isomorphisms.append(_rot_list ([complex(*pt) for pt,bl in points.value], i))
+                            isomorphisms.append(
+                                _rot_list([complex(*pt) for pt, bl in points.value], i)
+                            )
                     # Add mirrored rotations
                     mirrored = list(reversed(points.value))
                     reversed_bits = 0
-                    for pt,b in mirrored:
+                    for pt, b in mirrored:
                         reversed_bits = (reversed_bits << 1) | b
                     for i in range(n):
                         b = ((reversed_bits << i) & mask) | ((reversed_bits >> (n - i)))
                         if b == bits:
-                            isomorphisms.append(_rot_list ([complex(*pt) for pt,bl in mirrored], i))
+                            isomorphisms.append(
+                                _rot_list([complex(*pt) for pt, bl in mirrored], i)
+                            )
 
-            # Check each master against the first on in the list.
-            m0 = allNodeTypes[0]
-            for i,m1 in enumerate(allNodeTypes[1:]):
+            # m0idx should be the index of the first non-None item in allNodeTypes,
+            # else give it the first index of None, which is likely 0
+            m0idx = allNodeTypes.index(
+                next((x for x in allNodeTypes if x is not None), None)
+            )
+            # m0 is the first non-None item in allNodeTypes, or the first item if all are None
+            m0 = allNodeTypes[m0idx]
+            for i, m1 in enumerate(allNodeTypes[m0idx + 1 :]):
+                if m1 is None:
+                    continue
                 if len(m0) != len(m1):
                     add_problem(
                         glyph_name,
                         {
                             "type": "path_count",
-                            "master_1": names[0],
-                            "master_2": names[i + 1],
+                            "master_1": names[m0idx],
+                            "master_2": names[m0idx + i + 1],
                             "value_1": len(m0),
                             "value_2": len(m1),
                         },
@@ -262,8 +278,8 @@ def test(glyphsets, glyphs=None, names=None):
                             {
                                 "type": "node_count",
                                 "path": pathIx,
-                                "master_1": names[0],
-                                "master_2": names[i + 1],
+                                "master_1": names[m0idx],
+                                "master_2": names[m0idx + i + 1],
                                 "value_1": len(nodes1),
                                 "value_2": len(nodes2),
                             },
@@ -277,16 +293,24 @@ def test(glyphsets, glyphs=None, names=None):
                                     "type": "node_incompatibility",
                                     "path": pathIx,
                                     "node": nodeIx,
-                                    "master_1": names[0],
-                                    "master_2": names[i + 1],
+                                    "master_1": names[m0idx],
+                                    "master_2": names[m0idx + i + 1],
                                     "value_1": n1,
                                     "value_2": n2,
                                 },
                             )
                             continue
 
-            m0 = allVectors[0]
-            for i, m1 in enumerate(allVectors[1:]):
+            # m0idx should be the index of the first non-None item in allVectors,
+            # else give it the first index of None, which is likely 0
+            m0idx = allVectors.index(
+                next((x for x in allVectors if x is not None), None)
+            )
+            # m0 is the first non-None item in allVectors, or the first item if all are None
+            m0 = allVectors[m0idx]
+            for i, m1 in enumerate(allVectors[m0idx + 1 :]):
+                if m1 is None:
+                    continue
                 if len(m0) != len(m1):
                     # We already reported this
                     continue
@@ -296,39 +320,52 @@ def test(glyphsets, glyphs=None, names=None):
                 matching, matching_cost = min_cost_perfect_bipartite_matching(costs)
                 identity_matching = list(range(len(m0)))
                 identity_cost = sum(costs[i][i] for i in range(len(m0)))
-                if matching != identity_matching and matching_cost < identity_cost * .95:
+                if (
+                    matching != identity_matching
+                    and matching_cost < identity_cost * 0.95
+                ):
                     add_problem(
                         glyph_name,
                         {
                             "type": "contour_order",
-                            "master_1": names[0],
-                            "master_2": names[i + 1],
+                            "master_1": names[m0idx],
+                            "master_2": names[m0idx + i + 1],
                             "value_1": list(range(len(m0))),
                             "value_2": matching,
                         },
                     )
                     break
 
-            m0 = allContourIsomorphisms[0]
-            for i, m1 in enumerate(allContourIsomorphisms[1:]):
+            # m0idx should be the index of the first non-None item in allContourIsomorphisms,
+            # else give it the first index of None, which is likely 0
+            m0idx = allContourIsomorphisms.index(
+                next((x for x in allContourIsomorphisms if x is not None), None)
+            )
+            # m0 is the first non-None item in allContourIsomorphisms, or the first item if all are None
+            m0 = allContourIsomorphisms[m0idx]
+            for i, m1 in enumerate(allContourIsomorphisms[m0idx + 1 :]):
+                if m1 is None:
+                    continue
                 if len(m0) != len(m1):
                     # We already reported this
                     continue
                 if not m0:
                     continue
-                for ix,(contour0,contour1) in enumerate(zip(m0,m1)):
+                for ix, (contour0, contour1) in enumerate(zip(m0, m1)):
                     c0 = contour0[0]
-                    costs = [v for v in (_complex_vlen(_vdiff(c0, c1)) for c1 in contour1)]
+                    costs = [
+                        v for v in (_complex_vlen(_vdiff(c0, c1)) for c1 in contour1)
+                    ]
                     min_cost = min(costs)
                     first_cost = costs[0]
-                    if min_cost < first_cost * .95:
+                    if min_cost < first_cost * 0.95:
                         add_problem(
                             glyph_name,
                             {
                                 "type": "wrong_start_point",
                                 "contour": ix,
-                                "master_1": names[0],
-                                "master_2": names[i + 1],
+                                "master_1": names[m0idx],
+                                "master_2": names[m0idx + i + 1],
                             },
                         )
 
@@ -348,20 +385,37 @@ def main(args=None):
         "fonttools varLib.interpolatable",
         description=main.__doc__,
     )
+    parser.add_argument(
+        "--glyphs",
+        action="store",
+        help="Space-separate name of glyphs to check",
+    )
     parser.add_argument(
         "--json",
         action="store_true",
         help="Output report in JSON format",
     )
     parser.add_argument(
-        "inputs", metavar="FILE", type=str, nargs="+", help="Input TTF/UFO files"
+        "--quiet",
+        action="store_true",
+        help="Only exit with code 1 or 0, no output",
+    )
+    parser.add_argument(
+        "--ignore-missing",
+        action="store_true",
+        help="Will not report glyphs missing from sparse masters as errors",
+    )
+    parser.add_argument(
+        "inputs",
+        metavar="FILE",
+        type=str,
+        nargs="+",
+        help="Input a single DesignSpace/Glyphs file, or multiple TTF/UFO files",
     )
 
     args = parser.parse_args(args)
-    glyphs = None
-    # glyphs = ['uni08DB', 'uniFD76']
-    # glyphs = ['uni08DE', 'uni0034']
-    # glyphs = ['uni08DE', 'uni0034', 'uni0751', 'uni0753', 'uni0754', 'uni08A4', 'uni08A4.fina', 'uni08A5.fina']
+
+    glyphs = set(args.glyphs.split()) if args.glyphs else None
 
     from os.path import basename
 
@@ -369,31 +423,34 @@ def main(args=None):
     names = []
 
     if len(args.inputs) == 1:
-        if args.inputs[0].endswith('.designspace'):
+        if args.inputs[0].endswith(".designspace"):
             from fontTools.designspaceLib import DesignSpaceDocument
+
             designspace = DesignSpaceDocument.fromfile(args.inputs[0])
             args.inputs = [master.path for master in designspace.sources]
 
-        elif args.inputs[0].endswith('.glyphs'):
+        elif args.inputs[0].endswith(".glyphs"):
             from glyphsLib import GSFont, to_ufos
+
             gsfont = GSFont(args.inputs[0])
             fonts.extend(to_ufos(gsfont))
-            names = ['%s-%s' % (f.info.familyName, f.info.styleName) for f in fonts]
+            names = ["%s-%s" % (f.info.familyName, f.info.styleName) for f in fonts]
             args.inputs = []
 
-        elif args.inputs[0].endswith('.ttf'):
+        elif args.inputs[0].endswith(".ttf"):
             from fontTools.ttLib import TTFont
+
             font = TTFont(args.inputs[0])
-            if 'gvar' in font:
+            if "gvar" in font:
                 # Is variable font
-                gvar = font['gvar']
+                gvar = font["gvar"]
                 # Gather all "master" locations
                 locs = set()
                 for variations in gvar.variations.values():
                     for var in variations:
                         loc = []
-                        for tag,val in sorted(var.axes.items()):
-                            loc.append((tag,val[1]))
+                        for tag, val in sorted(var.axes.items()):
+                            loc.append((tag, val[1]))
                         locs.add(tuple(loc))
                 # Rebuild locs as dictionaries
                 new_locs = [{}]
@@ -401,7 +458,7 @@ def main(args=None):
                 for loc in sorted(locs, key=lambda v: (len(v), v)):
                     names.append(str(loc))
                     l = {}
-                    for tag,val in loc:
+                    for tag, val in loc:
                         l[tag] = val
                     new_locs.append(l)
                 locs = new_locs
@@ -413,82 +470,108 @@ def main(args=None):
 
                 args.inputs = []
 
-
     for filename in args.inputs:
         if filename.endswith(".ufo"):
             from fontTools.ufoLib import UFOReader
+
             fonts.append(UFOReader(filename))
         else:
             from fontTools.ttLib import TTFont
+
             fonts.append(TTFont(filename))
 
         names.append(basename(filename).rsplit(".", 1)[0])
 
-    if hasattr(fonts[0], 'getGlyphSet'):
-        glyphsets = [font.getGlyphSet() for font in fonts]
-    else:
-        glyphsets = fonts
+    glyphsets = []
+    for font in fonts:
+        if hasattr(font, "getGlyphSet"):
+            glyphset = font.getGlyphSet()
+        else:
+            glyphset = font
+        glyphsets.append({k: glyphset[k] for k in glyphset.keys()})
 
-    problems = test(glyphsets, glyphs=glyphs, names=names)
-    if args.json:
-        import json
+    if not glyphs:
+        glyphs = set([gn for glyphset in glyphsets for gn in glyphset.keys()])
 
-        print(json.dumps(problems))
-    else:
-        for glyph, glyph_problems in problems.items():
-            print(f"Glyph {glyph} was not compatible: ")
-            for p in glyph_problems:
-                if p["type"] == "missing":
-                    print("    Glyph was missing in master %s" % p["master"])
-                if p["type"] == "open_path":
-                    print("    Glyph has an open path in master %s" % p["master"])
-                if p["type"] == "path_count":
-                    print(
-                        "    Path count differs: %i in %s, %i in %s"
-                        % (p["value_1"], p["master_1"], p["value_2"], p["master_2"])
-                    )
-                if p["type"] == "node_count":
-                    print(
-                        "    Node count differs in path %i: %i in %s, %i in %s"
-                        % (
-                            p["path"],
-                            p["value_1"],
-                            p["master_1"],
-                            p["value_2"],
-                            p["master_2"],
+    for glyphset in glyphsets:
+        glyphSetGlyphNames = set(glyphset.keys())
+        diff = glyphs - glyphSetGlyphNames
+        if diff:
+            for gn in diff:
+                glyphset[gn] = None
+
+    problems = test(
+        glyphsets, glyphs=glyphs, names=names, ignore_missing=args.ignore_missing
+    )
+
+    if not args.quiet:
+        if args.json:
+            import json
+
+            print(json.dumps(problems))
+        else:
+            for glyph, glyph_problems in problems.items():
+                print(f"Glyph {glyph} was not compatible: ")
+                for p in glyph_problems:
+                    if p["type"] == "missing":
+                        print("    Glyph was missing in master %s" % p["master"])
+                    if p["type"] == "open_path":
+                        print("    Glyph has an open path in master %s" % p["master"])
+                    if p["type"] == "path_count":
+                        print(
+                            "    Path count differs: %i in %s, %i in %s"
+                            % (p["value_1"], p["master_1"], p["value_2"], p["master_2"])
                         )
-                    )
-                if p["type"] == "node_incompatibility":
-                    print(
-                        "    Node %o incompatible in path %i: %s in %s, %s in %s"
-                        % (
-                            p["node"],
-                            p["path"],
-                            p["value_1"],
-                            p["master_1"],
-                            p["value_2"],
-                            p["master_2"],
+                    if p["type"] == "node_count":
+                        print(
+                            "    Node count differs in path %i: %i in %s, %i in %s"
+                            % (
+                                p["path"],
+                                p["value_1"],
+                                p["master_1"],
+                                p["value_2"],
+                                p["master_2"],
+                            )
                         )
-                    )
-                if p["type"] == "contour_order":
-                    print(
-                        "    Contour order differs: %s in %s, %s in %s"
-                        % (
-                            p["value_1"],
-                            p["master_1"],
-                            p["value_2"],
-                            p["master_2"],
+                    if p["type"] == "node_incompatibility":
+                        print(
+                            "    Node %o incompatible in path %i: %s in %s, %s in %s"
+                            % (
+                                p["node"],
+                                p["path"],
+                                p["value_1"],
+                                p["master_1"],
+                                p["value_2"],
+                                p["master_2"],
+                            )
                         )
-                    )
-                if p["type"] == "wrong_start_point":
-                    print(
-                        "    Contour %d start point differs: %s, %s"
-                        % (
-                            p["contour"],
-                            p["master_1"],
-                            p["master_2"],
+                    if p["type"] == "contour_order":
+                        print(
+                            "    Contour order differs: %s in %s, %s in %s"
+                            % (
+                                p["value_1"],
+                                p["master_1"],
+                                p["value_2"],
+                                p["master_2"],
+                            )
+                        )
+                    if p["type"] == "wrong_start_point":
+                        print(
+                            "    Contour %d start point differs: %s, %s"
+                            % (
+                                p["contour"],
+                                p["master_1"],
+                                p["master_2"],
+                            )
+                        )
+                    if p["type"] == "math_error":
+                        print(
+                            "    Miscellaneous error in %s: %s"
+                            % (
+                                p["master"],
+                                p["error"],
+                            )
                         )
-                    )
     if problems:
         return problems
 
